==> sanctify-1.0.0.tar.gz: sanctify-1.0.0/sanctify/processor.py <==
from typing import Callable

import pandas as pd
from frozendict import frozendict
from loguru import logger

from sanctify.constants import (
    PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP,
    AbstractDataTypes,
    Constants,
    DefaultColumns,
    PrimitiveDataTypes,
)
from sanctify.transformer import Transformer
from sanctify.validator import Validator


def handle_additional_transformations_and_validations(
    data_type_name_for_column: str,
) -> list[Callable]:
    """
    Handles additional transformations and validations based on the data type name for a column.

    Args:
        data_type_name_for_column (str): Data type name for the column.

    Returns:
        list[Callable]: List of transformation functions to apply.
    """
    match data_type_name_for_column:
        case PrimitiveDataTypes.BOOLEAN.value:
            return [Transformer.parse_boolean]

        case PrimitiveDataTypes.FLOAT.value | AbstractDataTypes.DECIMAL.value:
            return [float]

        case PrimitiveDataTypes.INTEGER.value:
            return [int]

        case PrimitiveDataTypes.STRING.value | AbstractDataTypes.TEXT.value:
            return [str]

        case _:
            return []


def handle_data_type(data_type_schema: dict[str:dict], col_config: dict) -> tuple[list[Callable], list[Callable]]:
    """
    Handles data type transformations and validations for a column.

    Args:
        data_type_schema (dict[str: dict]): Data type schema containing data type configurations.
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.DATA_TYPE.value in col_config:
        data_type_name_for_column: str = col_config[Constants.DATA_TYPE.value]
        transformations += handle_additional_transformations_and_validations(
            data_type_name_for_column=data_type_name_for_column
        )
        data_type_config_for_column: dict = data_type_schema.get(data_type_name_for_column, {})
        transformations += data_type_config_for_column.get(Constants.TRANSFORMATIONS, [])
        validations += data_type_config_for_column.get(Constants.VALIDATIONS, [])

    return transformations, validations


def handle_date_order_tuple(col_config) -> tuple[list[Callable], list[Callable]]:
    """
    Handles date order transformations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.DATE_ORDER_TUPLE.value in col_config:
        date_order_tuple = col_config[Constants.DATE_ORDER_TUPLE.value]
        transformations = [
            (
                Transformer.parse_date_from_string,
                {Constants.DATE_ORDER_TUPLE.value: date_order_tuple},
            )
        ]

    return transformations, validations


def handle_comparison_operation(col_config) -> tuple[list[Callable], list[Callable]]:
    """
    Handles comparison operation validations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.COMPARISON_OPERATIONS.value in col_config:
        comparison_operations = col_config[Constants.COMPARISON_OPERATIONS.value]
        validations = [
            (
                Validator.validate_comparison_operations,
                {"comparison_operations": comparison_operations},
            )
        ]

    return transformations, validations


def handle_transformations(col_config) -> tuple[list[Callable], list[Callable]]:
    """
    Handles general transformations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.TRANSFORMATIONS.value in col_config:
        transformations = col_config[Constants.TRANSFORMATIONS.value]

    return transformations, validations


def handle_validations(col_config) -> tuple[list[Callable], list[Callable]]:
    """
    Handles validations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.VALIDATIONS.value in col_config:
        validations = col_config[Constants.VALIDATIONS.value]

    return transformations, validations


def handle_post_processing_dtype(col_config: dict, data_type_schema: dict) -> Callable | None:
    """
    Handles post validation transformations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
       Callable if int or float post validation transformation if applicable else returns None
    """
    dtype_in_col_config = PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP.get(
        col_config.get(Constants.POST_PROCESSING_DATA_TYPE.value, ""), None
    )
    if dtype_in_col_config is None:
        data_type_name_for_column: str = col_config.get(Constants.DATA_TYPE.value, "")
        data_type_config_for_column: dict = data_type_schema.get(data_type_name_for_column, {})
        return PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP.get(
            data_type_config_for_column.get(Constants.POST_PROCESSING_DATA_TYPE.value, ""),
            None,
        )
    else:
        return dtype_in_col_config


def apply_functions(value, functions: list[Callable | tuple[Callable, dict]]) -> pd.Series:
    """
    Applies a series of transformation and validation functions to a value.

    Args:
        value: The value to apply the functions to.
        functions (list[Callable]): List of transformation and validation functions.
        functions (tuple[Callable, dict]]): List of tuple of transformation and validation functions and additional kwargs.

    Returns:
        pd.Series: Series containing the modified value and error message (if any).
    """
    func_value = value
    error_message = None

    for function in functions:
        func = function
        kwargs = {}

        if isinstance(function, tuple):
            func = function[0]
            kwargs = function[1]

        logger.debug(f"Running {func.__name__ = } with {value = } | {kwargs = }")

        try:
            func_value = func(func_value, **kwargs)

        except Exception as err:
            logger.error(f"FAILED: {func.__name__ = } with {value = } | {kwargs = } | {(str(err)) = }")
            logger.error(f"{err = }")
            error_message = f"@validation_error@ : {str(err)} | {func_value}"
            func_value = None
            break

        logger.debug(f"Result: {func.__name__ = } with {value = } | {kwargs = } | {func_value = }")

    # Return the modified value and error message (if any)
    return pd.Series(
        [func_value, error_message],
        index=[DefaultColumns.VALUE_PLACEHOLDER.value, DefaultColumns.ERROR.value],
    )


def process_cleansed_df(
    df: pd.DataFrame,
    column_mapping_schema: frozendict = {},
    data_type_schema: frozendict = {},
) -> pd.DataFrame:
    """
    Processes the cleansed DataFrame by applying transformations and validations based on the column mapping and
    data type schema.

    Args:
        df (pd.DataFrame): The cleansed DataFrame.
        column_mapping_schema (frozendict): Column mapping schema.
        data_type_schema (frozendict): Data type schema.

    Returns:
        pd.DataFrame: The processed DataFrame.
    """
    if column_mapping_schema == {}:
        return df

    # Create an "Error" column in the DataFrame to track errors
    df[DefaultColumns.ERROR.value] = ""

    # Iterate over each column in column_mapping_schema
    for col_name, col_config in column_mapping_schema.items():
        logger.debug(f"{col_name = } | {col_config = }")

        # Check if col_name in column mapping schema exists in df
        col_name_in_df_columns = col_name in df.columns
        if col_name_in_df_columns is False:
            logger.error(f"Column({col_name}) with doesn't exist in the input dataframe | {col_config = }")
            # Continue to next column in the column mapping schema
            continue

        if col_config == {}:
            logger.error(f"Column({col_name})'s column config not provided in the input dataframe | {col_config = }")
            # Continue to next column in the column mapping schema
            continue  # Skip empty config items
        else:
            logger.debug(f"Parsing Column({col_name}) as str dtype | {col_config = }")
            df[col_name] = df[col_name].astype(str)  # Convert column rows to string data type

        transformations, validations = [], []

        d_type_transformations, d_type_validations = handle_data_type(data_type_schema, col_config)
        d_order_transformations, d_order_validations = handle_date_order_tuple(col_config)
        (
            comparison_op_transformations,
            comparison_op_validations,
        ) = handle_comparison_operation(col_config)
        col_transformations, _ = handle_transformations(col_config)
        _, col_validations = handle_validations(col_config)
        post_processing_dtype = handle_post_processing_dtype(col_config=col_config, data_type_schema=data_type_schema)

        transformations += (
            d_type_transformations + d_order_transformations + comparison_op_transformations + col_transformations
        )
        validations += d_type_validations + d_order_validations + comparison_op_validations + col_validations

        # Apply transformations and validations on the column
        logger.debug(f"Running {col_name}: {transformations = }")
        logger.debug(f"Running {col_name}: {validations = }")

        data_type_name_for_column: str = col_config.get(Constants.DATA_TYPE.value)
        primitive_data_type_classes = handle_additional_transformations_and_validations(
            data_type_name_for_column=data_type_name_for_column
        )
        applied_series = df[col_name].apply(apply_functions, functions=transformations + validations)

        # Update cell values
        successful_rows = applied_series[DefaultColumns.ERROR.value].isnull()
        df.loc[successful_rows, col_name] = applied_series.loc[successful_rows, DefaultColumns.VALUE_PLACEHOLDER.value]

        # Update "Error" column for errored rows
        errored_rows = applied_series[DefaultColumns.ERROR.value].notnull()
        if errored_rows.sum() > 0:
            df.loc[errored_rows, col_name] = applied_series.loc[errored_rows, DefaultColumns.VALUE_PLACEHOLDER.value]
            df.loc[errored_rows, DefaultColumns.ERROR.value] += (
                " | " + applied_series.loc[errored_rows, DefaultColumns.ERROR.value]
            )

        if len(primitive_data_type_classes) == 1:
            primitive_data_type_class = primitive_data_type_classes[0]
            if primitive_data_type_class in PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP:
                nullable_dtype = PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP[primitive_data_type_class]
                df[col_name] = pd.to_numeric(df[col_name], errors="coerce").astype(nullable_dtype)
                logger.debug(f"Converting the column {col_name} back to {nullable_dtype} data type")

        if post_processing_dtype is not None:
            logger.debug(f"Converting the column {col_name} to {post_processing_dtype} data type")
            df[col_name] = pd.to_numeric(df[col_name], errors="coerce").astype(post_processing_dtype)

    logger.debug(f"Final | {df = }")

    return df

==> sanctify-1.0.2.tar.gz: sanctify-1.0.2/sanctify/processor.py <==
from typing import Callable

import pandas as pd
from frozendict import frozendict
from loguru import logger

from sanctify.constants import (
    PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP,
    AbstractDataTypes,
    Constants,
    DefaultColumns,
    PrimitiveDataTypes,
)
from sanctify.transformer import Transformer
from sanctify.validator import Validator


def handle_additional_transformations_and_validations(
    data_type_name_for_column: str,
) -> list[Callable]:
    """
    Handles additional transformations and validations based on the data type name for a column.

    Args:
        data_type_name_for_column (str): Data type name for the column.

    Returns:
        list[Callable]: List of transformation functions to apply.
    """
    match data_type_name_for_column:
        case PrimitiveDataTypes.BOOLEAN.value:
            return [Transformer.parse_boolean]

        case PrimitiveDataTypes.FLOAT.value | AbstractDataTypes.DECIMAL.value:
            return [float]

        case PrimitiveDataTypes.INTEGER.value | AbstractDataTypes.NUMBER.value:
            return [int]

        case PrimitiveDataTypes.STRING.value | AbstractDataTypes.TEXT.value:
            return [str]

        case _:
            return []


def handle_data_type(data_type_schema: dict[str:dict], col_config: dict) -> tuple[list[Callable], list[Callable]]:
    """
    Handles data type transformations and validations for a column.

    Args:
        data_type_schema (dict[str: dict]): Data type schema containing data type configurations.
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.DATA_TYPE.value in col_config:
        data_type_name_for_column: str = col_config[Constants.DATA_TYPE.value]
        transformations += handle_additional_transformations_and_validations(
            data_type_name_for_column=data_type_name_for_column
        )
        data_type_config_for_column: dict = data_type_schema.get(data_type_name_for_column, {})
        transformations += data_type_config_for_column.get(Constants.TRANSFORMATIONS, [])
        validations += data_type_config_for_column.get(Constants.VALIDATIONS, [])

    return transformations, validations


def handle_date_order_tuple(col_config) -> tuple[list[Callable], list[Callable]]:
    """
    Handles date order transformations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.DATE_ORDER_TUPLE.value in col_config:
        date_order_tuple = col_config[Constants.DATE_ORDER_TUPLE.value]
        transformations = [
            (
                Transformer.parse_date_from_string,
                {Constants.DATE_ORDER_TUPLE.value: date_order_tuple},
            )
        ]

    return transformations, validations


def handle_comparison_operation(col_config) -> tuple[list[Callable], list[Callable]]:
    """
    Handles comparison operation validations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.COMPARISON_OPERATIONS.value in col_config:
        comparison_operations = col_config[Constants.COMPARISON_OPERATIONS.value]
        validations = [
            (
                Validator.validate_comparison_operations,
                {"comparison_operations": comparison_operations},
            )
        ]

    return transformations, validations


def handle_transformations(col_config) -> tuple[list[Callable], list[Callable]]:
    """
    Handles general transformations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.TRANSFORMATIONS.value in col_config:
        transformations = col_config[Constants.TRANSFORMATIONS.value]

    return transformations, validations


def handle_validations(col_config) -> tuple[list[Callable], list[Callable]]:
    """
    Handles validations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
        tuple[list[Callable], list[Callable]]: List of transformation functions and list of validation functions.
    """
    transformations, validations = [], []
    if Constants.VALIDATIONS.value in col_config:
        validations = col_config[Constants.VALIDATIONS.value]

    return transformations, validations


def handle_post_processing_dtype(col_config: dict, data_type_schema: dict) -> Callable | None:
    """
    Handles post validation transformations for a column.

    Args:
        col_config (dict): Column configuration.

    Returns:
       Callable if int or float post validation transformation if applicable else returns None
    """
    dtype_in_col_config = PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP.get(
        col_config.get(Constants.POST_PROCESSING_DATA_TYPE.value, ""), None
    )
    if dtype_in_col_config is None:
        data_type_name_for_column: str = col_config.get(Constants.DATA_TYPE.value, "")
        data_type_config_for_column: dict = data_type_schema.get(data_type_name_for_column, {})
        return PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP.get(
            data_type_config_for_column.get(Constants.POST_PROCESSING_DATA_TYPE.value, ""),
            None,
        )
    else:
        return dtype_in_col_config


def apply_functions(value, col_name: str, functions: list[Callable | tuple[Callable, dict]]) -> pd.Series:
    """
    Applies a series of transformation and validation functions to a value.

    Args:
        value: The value to apply the functions to.
        functions (list[Callable]): List of transformation and validation functions.
        functions (tuple[Callable, dict]]): List of tuple of transformation and validation functions and additional kwargs.

    Returns:
        pd.Series: Series containing the modified value and error message (if any).
    """
    func_value = value
    error_message = None

    for function in functions:
        func = function
        kwargs = {}

        if isinstance(function, tuple):
            func = function[0]
            kwargs = function[1]

        logger.debug(f"Running {func.__name__ = } with {value = } | {kwargs = }")

        try:
            func_value = func(func_value, **kwargs)

        except Exception as err:
            logger.error(f"FAILED: {func.__name__ = } with {value = } | {kwargs = } | {(str(err)) = }")
            logger.error(f"{err = }")
            error_message = f"{col_name} : {str(err)} | {func_value}"
            func_value = None
            break

        logger.debug(f"Result: {func.__name__ = } with {value = } | {kwargs = } | {func_value = }")

    # Return the modified value and error message (if any)
    return pd.Series(
        [func_value, error_message],
        index=[DefaultColumns.VALUE_PLACEHOLDER.value, DefaultColumns.ERROR.value],
    )


def process_cleansed_df(
    df: pd.DataFrame,
    column_mapping_schema: frozendict = {},
    data_type_schema: frozendict = {},
) -> pd.DataFrame:
    """
    Processes the cleansed DataFrame by applying transformations and validations based on the column mapping and
    data type schema.

    Args:
        df (pd.DataFrame): The cleansed DataFrame.
        column_mapping_schema (frozendict): Column mapping schema.
        data_type_schema (frozendict): Data type schema.

    Returns:
        pd.DataFrame: The processed DataFrame.
    """
    if column_mapping_schema == {}:
        return df

    # Create an "Error" column in the DataFrame to track errors
    df[DefaultColumns.ERROR.value] = ""

    # Iterate over each column in column_mapping_schema
    for col_name, col_config in column_mapping_schema.items():
        logger.debug(f"{col_name = } | {col_config = }")

        # Check if col_name in column mapping schema exists in df
        col_name_in_df_columns = col_name in df.columns
        if col_name_in_df_columns is False:
            logger.error(f"Column({col_name}) with doesn't exist in the input dataframe | {col_config = }")
            # Continue to next column in the column mapping schema
            continue

        if col_config == {}:
            logger.error(f"Column({col_name})'s column config not provided in the input dataframe | {col_config = }")
            # Continue to next column in the column mapping schema
            continue  # Skip empty config items
        else:
            logger.debug(f"Parsing Column({col_name}) as str dtype | {col_config = }")
            df[col_name] = df[col_name].astype(str)  # Convert column rows to string data type

        transformations, validations = [], []

        d_type_transformations, d_type_validations = handle_data_type(data_type_schema, col_config)
        d_order_transformations, d_order_validations = handle_date_order_tuple(col_config)
        (
            comparison_op_transformations,
            comparison_op_validations,
        ) = handle_comparison_operation(col_config)
        col_transformations, _ = handle_transformations(col_config)
        _, col_validations = handle_validations(col_config)
        post_processing_dtype = handle_post_processing_dtype(col_config=col_config, data_type_schema=data_type_schema)

        transformations += (
            d_type_transformations + d_order_transformations + comparison_op_transformations + col_transformations
        )
        validations += d_type_validations + d_order_validations + comparison_op_validations + col_validations

        # Apply transformations and validations on the column
        logger.debug(f"Running {col_name}: {transformations = }")
        logger.debug(f"Running {col_name}: {validations = }")

        data_type_name_for_column: str = col_config.get(Constants.DATA_TYPE.value)
        primitive_data_type_classes = handle_additional_transformations_and_validations(
            data_type_name_for_column=data_type_name_for_column
        )
        applied_series = df[col_name].apply(apply_functions, col_name=col_name, functions=transformations + validations)

        # Update cell values
        successful_rows = applied_series[DefaultColumns.ERROR.value].isnull()
        df.loc[successful_rows, col_name] = applied_series.loc[successful_rows, DefaultColumns.VALUE_PLACEHOLDER.value]

        # Update "Error" column for errored rows
        errored_rows = applied_series[DefaultColumns.ERROR.value].notnull()
        if errored_rows.sum() > 0:
            df.loc[errored_rows, col_name] = applied_series.loc[errored_rows, DefaultColumns.VALUE_PLACEHOLDER.value]
            df.loc[errored_rows, DefaultColumns.ERROR.value] += (
                " | " + applied_series.loc[errored_rows, DefaultColumns.ERROR.value]
            )

        if len(primitive_data_type_classes) == 1:
            primitive_data_type_class = primitive_data_type_classes[0]
            if primitive_data_type_class in PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP:
                nullable_dtype = PRIMITIVE_TO_NULLABLE_DATA_TYPE_MAP[primitive_data_type_class]
                df[col_name] = pd.to_numeric(df[col_name], errors="coerce").astype(nullable_dtype)
                logger.debug(f"Converting the column {col_name} back to {nullable_dtype} data type")

        if post_processing_dtype is not None:
            logger.debug(f"Converting the column {col_name} to {post_processing_dtype} data type")
            df[col_name] = pd.to_numeric(df[col_name], errors="coerce").astype(post_processing_dtype)

    logger.debug(f"Final | {df = }")

    return df
